# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoServerSessionList and X2GoServerSessionInfo classes - data handling for
X2Go server sessions.

This backend handles X2Go server implementations that respond with session infos
via server-side PLAIN text output.

"""

from __future__ import print_function
__NAME__ = 'x2goserversessioninfo-pylib'

__package__ = 'x2go.backends.info'
__name__    = 'x2go.backends.info.plain'

# modules
import types
import re

class X2GoServerSessionInfo(object):
    """\
    :class:`x2go.backends.info.plain.X2GoServerSessionInfo` is used to store all information
    that is retrieved from the connected X2Go server on
    ``X2GoTerminalSession.start()`` resp. ``X2GoTerminalSession.resume()``.

    """
    def __init__(self):
        self.initialized = False
        self.protected = False
        self.session_type_cached = None

    def __str__(self):
        return self.name

    def __repr__(self):
        result = 'X2GoServerSessionInfo('
        for p in dir(self):
            if '__' in p or not p in self.__dict__: continue
            result += p + '=' + str(self.__dict__[p]) +','
        return result.strip(',') + ')'

    def _parse_x2golistsessions_line(self, x2go_output):
        """\
        Parse a single line of X2Go's listsessions output.

        :param x2go_output: output from ,,x2golistsessions'' command (as list of strings/lines)
        :type x2go_output: ``list``

        """
        try:
            l = x2go_output.split("|")
            self.agent_pid = int(l[0])
            self.name = l[1]
            self.display = int(l[2])
            self.hostname = l[3]
            self.status = l[4]
            # TODO: turn into datetime object
            self.date_created = l[5]
            self.cookie = l[6]
            self.graphics_port = int(l[8])
            self.snd_port = int(l[9])
            # TODO: turn into datetime object
            self.date_suspended = l[10]
            self.username = l[11]
            self.sshfs_port = int(l[13])
            self.local_container = ''
            self.initialized = True
        except IndexError as e:
            # DEBUGGING CODE
            raise e
        except ValueError as e:
            # DEBUGGING CODE
            raise e

        # retrieve Telekinesis ports from list of sessions...
        try:
            self.tekictrl_port = int(l[14])
        except (IndexError, ValueError) as e:
            self.tekictrl_port = -1
        try:
            self.tekidata_port = int(l[15])
        except (IndexError, ValueError) as e:
            self.tekidata_port = -1

    def is_published_applications_provider(self):
        """\
        Detect from session info if this session is a published applications provider.


        :returns: returns ``True`` if this session is a published applications provider

        :rtype: ``bool`` or ``None`` (if not initialized)

        """
        if self.initialized:
            return bool(re.match('.*_stRPUBLISHED_.*', self.name))
        else:
            return None

    def is_running(self):
        """\
        Is this session running?


        :returns: ``True`` if the session is running, ``False`` otherwise

        :rtype: ``bool`` or ``None`` (if not initialized)

        """
        if self.initialized:
            return self.status == 'R'
        else:
            return None

    def get_session_type(self):
        """\
        Get the session type (i.e. 'D', 'K', 'R', 'S' or 'P').


        :returns: session type

        :rtype: ``str`` or ``None`` (if not initialized)

        """
        if self.session_type_cached is not None:
            return self.session_type_cached
        if self.initialized:
            cmd = self.name.split('_')[1]
            session_type = cmd[2]
            if session_type == 'R' and self.is_published_applications_provider():
                session_type = 'P'
            self.session_type_cached = session_type
            return session_type
        else:
            return None

    def get_share_mode(self):
        """\
        Get the share mode of a shadow session.


        :returns: share mode (0: view-only, 1: full access), ``None`` when used for non-desktop-sharing sessions

        :rtype: ``str`` or ``None`` (if not initialized)

        """
        if self.initialized:
            share_mode = None
            cmd = self.name.split('_')[1]
            session_type = cmd[2]
            if session_type == 'S':
                share_mode = cmd[3]
            return share_mode
        else:
            return None

    def is_suspended(self):
        """\
        Is this session suspended?


        :returns: ``True`` if the session is suspended, ``False`` otherwise

        :rtype: ``bool`` or ``None`` (if not initialized)

        """
        if self.initialized:
            return self.status == 'S'
        else:
            return None

    def is_desktop_session(self):
        """\
        Is this session a desktop session?


        :returns: ``True`` if this session is a desktop session, ``False`` otherwise

        :rtype: ``bool`` or ``None`` (if not initialized)

        """
        if self.initialized:
            return self.get_session_type() in ('D', 'K')
        else:
            return None

    def is_kdrive_session(self):
        """\
        Is this session a KDrive based desktop session?


        :returns: ``True`` if this session is a KDrive based desktop session, ``False`` otherwise

        :rtype: ``bool``

        """
        if self.initialized:
            return self.get_session_type() == 'K'
        else:
           return None

    def is_nx3_session(self):
        """\
        Is this session an NXv3 based desktop/rootless/shadow session?


        :returns: ``True`` if this session is a NXv3 based desktop, rootless or shadow session, ``False`` otherwise

        :rtype: ``bool``

        """
        if self.initialized:
            return self.get_session_type() in ('D', 'S', 'R', 'P')
        else:
            return None

    def is_kdrive_session(self):
        """\
        Is this session a KDrive based desktop session?


        :returns: ``True`` if this session is a KDrive based desktop session, ``False`` otherwise

        :rtype: ``bool``

        """
        return self.get_session_type() == 'K'

    def is_nx3_session(self):
        """\
        Is this session an NXv3 based desktop/rootless session?


        :returns: ``True`` if this session is a NXv3 based desktop/rootless session, ``False`` otherwise

        :rtype: ``bool``

        """
        return self.get_session_type() in ('D','R','P')

    def _parse_x2gostartagent_output(self, x2go_output):
        """\
        Parse x2gostartagent output.

        :param x2go_output: output from ,,x2gostartagent'' command (as list of strings/lines)
        :type x2go_output: ``list``

        """
        try:
            l = x2go_output.split("\n")
            self.name = l[3]
            self.cookie = l[1]
            self.agent_pid = int(l[2])
            self.display = int(l[0])
            self.graphics_port = int(l[4])
            self.snd_port = int(l[5])
            self.sshfs_port = int(l[6])
            self.username = ''
            self.hostname = ''
            # TODO: we have to see how we fill these fields here...
            self.date_created = ''
            self.date_suspended = ''
            # TODO: presume session is running after x2gostartagent, this could be better
            self.status = 'R'
            self.local_container = ''
            self.remote_container = ''
            self.initialized = True
        except IndexError as e:
            # DEBUGGING CODE
            raise e
        except ValueError as e:
            # DEBUGGING CODE
            raise e

        # retrieve Telekinesis ports from x2gostartagent output
        try:
            self.tekictrl_port = int(l[7])
        except (IndexError, ValueError) as e:
            self.tekictrl_port = -1
        try:
            self.tekidata_port = int(l[8])
        except (IndexError, ValueError) as e:
            self.tekidata_port = -1

    def initialize(self, x2go_output, username='', hostname='', local_container='', remote_container=''):
        """\
        Setup a a session info data block, includes parsing of X2Go server's ``x2gostartagent`` stdout values.

        :param x2go_output: X2Go server's ``x2gostartagent`` command output, each value
            separated by a newline character.
        :type x2go_output: str
        :param username: session user name (Default value = '')
        :type username: str
        :param hostname: hostname of X2Go server (Default value = '')
        :type hostname: str
        :param local_container: X2Go client session directory for config files, cache and session logs (Default value = '')
        :type local_container: str
        :param remote_container: X2Go server session directory for config files, cache and session logs (Default value = '')
        :type remote_container: str

        """
        self.protect()
        self._parse_x2gostartagent_output(x2go_output)
        self.username = username
        self.hostname = hostname
        self.local_container = local_container
        self.remote_container = remote_container
        self.initialized = True

    def protect(self):
        """\
        Write-protect this session info data structure.


        """
        self.protected = True

    def unprotect(self):
        """\
        Remove write-protection from this session info data structure.


        """
        self.protected = False

    def is_protected(self):
        """\
        Check if this session info data structure is Write-protected.


        """
        return self.protected

    def is_initialized(self):
        """\
        Check if this session info data structure has been initialized.


        """
        return self.initialized

    def get_status(self):
        """\
        Retrieve the session's status from this session info data structure.


        :returns: session status

        :rtype: ``str`` or ``None`` (if not initalized)

        """
        if self.initialized:
            return self.status
        else:
            return None

    def clear(self):
        """\
        Clear all properties of a :class:`x2go.backends.info.plain.X2GoServerSessionInfo` object.


        """
        self.name = ''
        self.cookie = ''
        self.agent_pid = ''
        self.display = ''
        self.graphics_port = ''
        self.snd_port = ''
        self.sshfs_port = ''
        self.tekictrl_port = ''
        self.tekidata_port = ''
        self.username = ''
        self.hostname = ''
        self.date_created = ''
        self.date_suspended = ''
        self.status = ''
        self.local_container = ''
        self.remote_container = ''
        self.protected = False
        self.initialized = False
        self.session_type_cached = None

    def update(self, session_info):
        """\
        Update all properties of a :class:`x2go.backends.info.plain.X2GoServerSessionInfo` object.

        :param session_info: a provided session info data structure
        :type session_info: ``X2GoServerSessionInfo*``

        """
        if type(session_info) == type(self):
            for prop in ('graphics_port', 'snd_port', 'sshfs_port', 'tekictrl_port', 'tekidata_port', 'date_suspended', 'status', ):
                if hasattr(session_info, prop):
                    _new = getattr(session_info, prop)
                    _current = getattr(self, prop)
                    if _new != _current:
                        setattr(self, prop, _new)

    def __init__(self):
        """\
        Class constructor, identical to :func:`clear()` method.

        """
        self.clear()


class X2GoServerSessionList(object):
    """\
    :class:`x2go.backends.info.plain.X2GoServerSessionList` is used to store all information
    that is retrieved from a connected X2Go server on a
    ``X2GoControlSession.list_sessions()`` call.


    """
    def __init__(self, x2go_output=None, info_backend=X2GoServerSessionInfo):
        """\
        :param x2go_output: X2Go server's ``x2golistsessions`` command output, each
            session separated by a newline character. Session values are separated
            by Unix Pipe Symbols ('|')
        :type x2go_output: str
        :param info_backend: the session info backend to use
        :type info_backend: ``X2GoServerSessionInfo*``

        """
        self.sessions = {}
        if x2go_output is not None:
            lines = x2go_output.split("\n")
            for line in lines:
                if not line:
                    continue
                s_info = info_backend()
                s_info._parse_x2golistsessions_line(line)
                self.sessions[s_info.name] = s_info

    def __call__(self):
        return self.sessions

    def get_sessions(self):
        """\
        Get the complete sessions property.

        :return: get this instance's dictionary of sessions
        :rtype: ``dict``

        """
        return self.sessions

    def set_sessions(self, sessions):
        """\
        Set the sessions property directly by parsing a complete data structure.

        :param sessions: set this instance's list of sessions directly
        :type sessions: ``dict``

        """
        self.sessions = sessions

    def get_session_info(self, session_name):
        """\
        Retrieve the session information for ``<session_name>``.

        :param session_name: the queried session name
        :type session_name: ``str``
        :returns: the session info of ``<session_name>``
        :rtype: ``X2GoServerSessionInfo*`` or ``None``

        """
        try:
            return self.sessions[session_name]
        except KeyError:
            return None

    def get_session_with(self, property_name, value, hostname=None):
        """\
        Find session with a given display number on a given host.

        :param property_name: match a session based on this property name
        :type property_name: ``str``
        :param value: the resulting session has to match this value for ``<property_name>``
        :type value: ``str``
        :param hostname: the result has to match this hostname (Default value = None)
        :type hostname: ``str``

        """
        if property_name == 'display':
            value = value.lstrip(':')
            if '.' in value: value = value.split('.')[0]

        for session in list(self.sessions.values()):
            try:
                if str(getattr(session, property_name)) == str(value):
                    if hostname is None or session.hostname == hostname:
                        return session
            except AttributeError:
                pass



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
:class:`x2go.printing.X2GoPrintQueue` sets up a thread that listens for incoming print jobs.

For each incoming print job in an X2Go session's spool directory an
individual thread is started (:class:`x2go.printqueue.X2GoPrintJob`) that handles the processing
of the incoming print job.

"""
__NAME__ = 'x2goprintqueue-pylib'

__package__ = 'x2go'
__name_     = 'x2go.printqueue'

# modules
import os
import copy
import threading
import gevent

# Python X2Go modules
from . import defaults
from . import utils
from . import log

from .defaults import X2GO_PRINTING_FILENAME as _X2GO_PRINTING_FILENAME
from .defaults import BACKENDS as _BACKENDS

class X2GoPrintQueue(threading.Thread):
    """\
    If X2Go printing is supported in a particular :class:`x2go.session.X2GoSession` instance
    this class provides a sub-thread for handling incoming X2Go print jobs.


    """
    print_action = None

    spooldir = None
    active_jobs = {}
    job_history = []

    def __init__(self,
                 profile_name='UNKNOWN',
                 session_name='UNKNOWN',
                 spool_dir=None,
                 print_action=None,
                 print_action_args={},
                 client_instance=None,
                 printing_backend=_BACKENDS['X2GoClientPrinting']['default'],
                 logger=None,
                 loglevel=log.loglevel_DEFAULT):
        """\
        :param profile_name: name of the session profile this print queue belongs to
        :type profile_name: ``str``
        :param spool_dir: local spool directory for incoming print job files
        :type spool_dir: ``str``
        :param print_action: name or instance of either of the possible X2Go print action classes
        :type print_action: ``str`` or instance
        :param print_action_args: depending of the chosen ``print_action`` this dictionary may contain different
            values; the ``print_action_args`` will be passed on to the X2Go print action instance constructor, so
            refer to either of these: :class:`x2go.printactions.X2GoPrintActionPDFVIEW`, :class:`x2go.printactions.X2GoPrintActionPRINT` et al.
        :param client_instance: the underlying :class:`x2go.client.X2GoClient` instance
        :type client_instance: ``obj``
        :param printing_backend: the client printing configuration backend class
        :type printing_backend: ``obj``
        :param logger: you can pass an :class:`x2go.log.X2GoLogger` object to the
            :class:`x2go.printing.X2GoPrintQueue` constructor
        :type logger: ``obj``
        :param loglevel: if no :class:`x2go.log.X2GoLogger` object has been supplied a new one will be
            constructed with the given loglevel
        :type loglevel: ``int``

        """
        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self.profile_name = profile_name
        self.session_name = session_name
        self.spool_dir = spool_dir
        if self.spool_dir: self.spool_dir = os.path.normpath(self.spool_dir)
        self.client_instance = client_instance
        self.client_rootdir = client_instance.get_client_rootdir()
        self.printing_backend = utils._get_backend_class(printing_backend, "X2GoClientPrinting")
        if print_action is not None:
            self.set_print_action(print_action, client_instance=self.client_instance, logger=logger, **print_action_args)
        threading.Thread.__init__(self)
        self.daemon = True
        self._accept_jobs = True

    def __del__(self):
        """\
        Class destructor.

        """
        self.stop_thread()

    def pause(self):
        """\
        Prevent acceptance of new incoming print jobs. The processing of print jobs that
        are currently still active will be completed, though.


        """
        if self._accept_jobs == True:
            self._accept_jobs = False
            self.logger('paused thread: %s' % repr(self), loglevel=log.loglevel_DEBUG)

    def resume(self):
        """\
        Resume operation of the X2Go print spooler and continue accepting new incoming
        print jobs.


        """
        if self._accept_jobs == False:
            self._accept_jobs = True
            self.logger('resumed thread: %s' % repr(self), loglevel=log.loglevel_DEBUG)

    def stop_thread(self):
        """\
        Stops this :class:`x2go.printing.X2GoPrintQueue` thread completely.


        """
        self.pause()
        self._keepalive = False
        self.logger('stopping thread: %s' % repr(self), loglevel=log.loglevel_DEBUG)

    @property
    def _incoming_print_jobs(self):

        if os.path.exists(self.spool_dir):
            l = os.listdir(self.spool_dir)
            job_files = [ jf for jf in l if jf.endswith('.ready') ]
            jobs = []
            for _job_file in job_files:
                _job_file_handle = open(os.path.join(self.spool_dir, _job_file), 'r')
                content = _job_file_handle.read()
                try:
                    (pdf_filename, job_title) = content.split('\n')[0:2]
                except ValueError:
                    pdf_filename = content
                    job_title = 'X2Go Print Job'
                _job_file_handle.close()
                jobs.append((_job_file, pdf_filename, job_title))
            return [ j for j in jobs if j[1] not in list(self.active_jobs.keys()) ]
        else:
            return []

    def set_print_action(self, print_action, **kwargs):
        """\
        Modify the print action of this :class:`x2go.printing.X2GoPrintQueue` thread during runtime. The
        change of print action will be valid for the next incoming print job.

        As kwargs you can pass arguments for the print action class to be set. Refer
        to the class descriptions of :class:`x2go.printactions.X2GoPrintActionDIALOG`, :class:`x2go.printactions.X2GoPrintActionPDFVIEW`,
        :class:`x2go.printactions.X2GoPrintActionPRINT`, etc.

        :param print_action: new print action to be valid for incoming print jobs
        :type print_action: ``str`` or ``class``
        :param kwargs: extra options for the specified print action
        :type kwargs: ``dict``

        """
        if print_action in list(defaults.X2GO_PRINT_ACTIONS.keys()):
            print_action = defaults.X2GO_PRINT_ACTIONS[print_action]

        if print_action in list(defaults.X2GO_PRINT_ACTIONS.values()):
            self.print_action = eval ('printactions.%s(**kwargs)' % print_action)

    def run(self):
        """\
        Start this :class:`x2go.printing.X2GoPrintQueue` thread...


        """
        self.logger('starting print queue thread: %s' % repr(self), loglevel=log.loglevel_DEBUG)

        self._keepalive = True
        while self._keepalive:

            while self._accept_jobs:

                if self._incoming_print_jobs:

                    for _job in self._incoming_print_jobs:
                        self.logger('processing incoming X2Go print job: %s' % _job[1], loglevel=log.loglevel_NOTICE)
                        _new_printjob_thread = X2GoPrintJob(target=x2go_printjob_handler,
                                                            kwargs={
                                                            'job_file': _job[0],
                                                            'pdf_file': _job[1],
                                                            'job_title': _job[2],
                                                            'print_action': self.print_action,
                                                            'parent_thread': self,
                                                            'logger': self.logger,
                                                          }
                                                  )
                        self.active_jobs['%s' % _job[1]] = _new_printjob_thread
                        _new_printjob_thread.start()

                gevent.sleep(3)

            gevent.sleep(1)


def x2go_printjob_handler(job_file=None, pdf_file=None, job_title=None, print_action=None, parent_thread=None, logger=None, ):
    """\
    This function is called as a handler function for each incoming X2Go print job
    represented by the class :class:`x2go.printqueue.X2GoPrintJob`.

    The handler function will (re-)read the »printing« configuration file (if no
    explicit ``print_action`` is passed to this function...). It then will
    execute the ``<print_action>.do_print()`` command.

    :param job_file: file name of this print job's Job File (Default value = None)
    :type job_file: ``str``
    :param pdf_file: PDF file name as placed in to the X2Go spool directory (Default value = None)
    :type pdf_file: ``str``
    :param job_title: human readable print job title (Default value = None)
    :type job_title: ``str``
    :param print_action: an instance of either of the possible ``X2GoPrintActionXXX`` classes (Default value = None)
    :type print_action: ``X2GoPrintActionXXX`` nstance
    :param parent_thread: the :class:`x2go.printing.X2GoPrintQueue` thread that actually created this handler's :class:`x2go.printqueue.X2GoPrintJob` instance (Default value = None)
    :type parent_thread: ``obj``
    :param logger: the :class:`x2go.printing.X2GoPrintQueue`'s logging instance (Default value = None)
    :type logger: ``obj``

    """
    if print_action is None:
        if parent_thread.client_instance is not None and parent_thread.client_instance.has_custom_client_rootdir:
            _printing = parent_thread.printing_backend(config_files=[os.path.join(parent_thread.client_instance.get_client_rootdir(), _X2GO_PRINTING_FILENAME)],
                                                       client_instance=parent_thread.client_instance,
                                                       logger=logger
                                                      )
        else:
            _printing = parent_thread.printing_backend(client_instance=parent_thread.client_instance,
                                                       logger=logger
                                                      )

        print_action = _printing.print_action
    print_action.profile_name = parent_thread.profile_name
    print_action.session_name = parent_thread.session_name

    logger('action for printing is: %s' % print_action, loglevel=log.loglevel_DEBUG)
    print_action.do_print(pdf_file=os.path.normpath(os.path.join(parent_thread.spool_dir, pdf_file)),
                          job_title=job_title,
                          spool_dir=parent_thread.spool_dir,
                         )

    logger('removing print job files for %s' % pdf_file, loglevel=log.loglevel_DEBUG)

    utils.patiently_remove_file(parent_thread.spool_dir, job_file)
    logger('removed print job file %s' % job_file, loglevel=log.loglevel_DEBUG)
    utils.patiently_remove_file(parent_thread.spool_dir, pdf_file)
    logger('removed print pdf file %s' % pdf_file, loglevel=log.loglevel_DEBUG)

    del parent_thread.active_jobs['%s' % pdf_file]
    parent_thread.job_history.append(pdf_file)

    # in case we print a lot we do not want to risk an endlessly growing
    # print job history
    if len(parent_thread.job_history) > 100:
        parent_thread.job_history = parent_thread.job_history[-100:]


class X2GoPrintJob(threading.Thread):
    """\
    For each X2Go print job we create a sub-thread that let's
    the print job be processed in the background.

    As a handler for this class the function :func:`x2go_printjob_handler()`
    is used.


    """
    def __init__(self, **kwargs):
        """\
        Construct the X2Go print job thread...

        All parameters (**kwargs) are passed through to the constructor
        of ``threading.Thread()``.

        """
        threading.Thread.__init__(self, **kwargs)
        self.daemon = True



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
:class:`x2go.backends.profiles.httpbroker.X2GoSessionProfiles` class - managing X2Go Client session profiles obtain from an HTTP based X2Go Session Broker.

:class:`x2go.backends.profiles.httpbroker.X2GoSessionProfiles` is a public API class. Use this class in your Python X2Go based
applications.

"""
from __future__ import print_function
__NAME__ = 'x2gosessionprofiles-pylib'

__package__ = 'x2go.backends.profiles'
__name__ = 'x2go.backends.profiles.httpbroker'

import re
import requests
import urllib3.exceptions
import copy
import types
import time
try: import simplejson as json
except ImportError: import json

# Python X2Go modules
from x2go.defaults import X2GO_SESSIONPROFILE_DEFAULTS as _X2GO_SESSIONPROFILE_DEFAULTS
from x2go.defaults import CURRENT_LOCAL_USER as _CURRENT_LOCAL_USER
import x2go.backends.profiles.base as base
import x2go.log as log
from x2go.utils import genkeypair
import x2go.x2go_exceptions

class X2GoSessionProfiles(base.X2GoSessionProfiles):

    defaultSessionProfile = copy.deepcopy(_X2GO_SESSIONPROFILE_DEFAULTS)

    def __init__(self, session_profile_defaults=None,
                 broker_url="http://localhost:8080/json/",
                 broker_username=None,
                 broker_password=None,
                 logger=None, loglevel=log.loglevel_DEFAULT,
                 **kwargs):
        """\
        Retrieve X2Go session profiles from a HTTP(S) session broker.

        :param session_profile_defaults: a default session profile
        :type session_profile_defaults: ``dict``
        :param broker_url: URL for accessing the X2Go Session Broker
        :type broker_url: ``str``
        :param broker_password: use this password for authentication against the X2Go Session Broker (avoid
            password string in the ``broker_URL`` parameter is highly recommended)
        :type broker_password: ``str``
        :param logger: you can pass an :class:`x2go.log.X2GoLogger` object to the
            :class:`x2go.backends.profiles.httpbroker.X2GoSessionProfiles` constructor
        :type logger: :class:`x2go.log.X2GoLogger` instance
        :param loglevel: if no :class:`x2go.log.X2GoLogger` object has been supplied a new one will be
            constructed with the given loglevel
        :type loglevel: ``int``

        """
        if broker_url.upper() != "HTTP":
            match = re.match('^(?P<protocol>(http(|s)))://(|(?P<user>[a-zA-Z0-9_\.-]+)(|:(?P<password>.*))@)(?P<hostname>[a-zA-Z0-9\.-]+)(|:(?P<port>[0-9]+))($|/(?P<path>.*)$)', broker_url)
            p = match.groupdict()
            if p['user']:
                self.broker_username = p['user']
            else:
                self.broker_username = broker_username
            if p['password']:
                self.broker_password = p['password']
            elif broker_password:
                self.broker_password = broker_password
            else:
                self.broker_password = None

            # fine-tune the URL
            p['path'] = "/{path}".format(**p)
            if p['port'] is not None:
                p['port'] = ":{port}".format(**p)
            else:
                p['port'] = ''

            self.broker_url = "{protocol}://{hostname}{port}{path}".format(**p)

        else:
            self.broker_username = broker_username
            self.broker_password = broker_password
            self.broker_url = broker_url

        self.broker_noauth = False
        self.broker_authid = None
        self._broker_profile_cache = {}
        self._mutable_profile_ids = None
        self._broker_auth_successful = None

        self._broker_type = "http"

        base.X2GoSessionProfiles.__init__(self, session_profile_defaults=session_profile_defaults, logger=logger, loglevel=loglevel)
        if self.broker_url != "HTTP":
            self.logger("Using session broker at URL: %s" % self.broker_url, log.loglevel_NOTICE)

        # for broker based autologin, we have to be able to provide public/private key pair
        self.broker_my_pubkey, self.broker_my_privkey = genkeypair(local_username=_CURRENT_LOCAL_USER, client_address='127.0.0.1')

    def get_broker_noauth(self):
        """\
        Accessor for the class's ``broker_noauth`` property.


        :returns: ``True`` if the broker probably does not expect authentication.

        :rtype: ``bool``

        """
        return self.broker_noauth

    def get_broker_username(self):
        """\
        Accessor for the class's ``broker_username`` property.


        :returns: the username used for authentication against the session broker URL

        :rtype: ``str``

        """
        return self.broker_username

    def get_broker_url(self):
        """\
        Accessor for the class's ``broker_url`` property.


        :returns: the session broker URL that was used at broker session instantiation

        :rtype: ``str``

        """
        return self.broker_url

    def set_broker_url(self, broker_url):
        """\
        Mutator for the class's ``broker_url`` property.

        :param broker_url: A new broker URL to use with this instance. Format is
            ``<protocol>://<hostname>:<port>/<path>`` (where protocol has to be ``http``
            or ``https``.
        :type broker_url: ``str``
        :returns: the session broker URL that was used at broker session instantiation
        :rtype: ``str``

        """
        self.broker_url = broker_url

    def get_broker_type(self):
        """\
        Accessor of the class's {_broker_type} property.


        :returns: either ``http`` or ``https``.

        :rtype: ``str``

        """
        return self._broker_type

    def broker_simpleauth(self, broker_username, broker_password):
        """\
        Attempt a username / password authentication against the instance's
        broker URL.

        :param broker_username: username to use for authentication
        :type broker_username: ``str``
        :param broker_password: password to use for authentication
        :type broker_password: ``str``
        :returns: ``True`` if authentication has been successful
        :rtype: ``bool``
        :raises X2GoBrokerConnectionException: Raised on any kind of connection /
            authentication failure.

        """
        if self.broker_url is not None:
            request_data = {
                'user': broker_username or '',
            }
            if self.broker_authid is not None:
                request_data['authid'] = self.broker_authid
                self.logger("Sending request to broker: user: {user}, authid: {authid}".format(**request_data), log.loglevel_DEBUG)
            else:
                if broker_password:
                    request_data['password'] = "<hidden>"
                else:
                    request_data['password'] = "<EMPTY>"
                self.logger("Sending request to broker: user: {user}, password: {password}".format(**request_data), log.loglevel_DEBUG)
                request_data['password'] = broker_password or ''
            try:
                r = requests.post(self.broker_url, data=request_data)
            except (requests.exceptions.ConnectionError, requests.exceptions.MissingSchema, urllib3.exceptions.LocationParseError):
                raise x2go.x2go_exceptions.X2GoBrokerConnectionException('Failed to connect to URL %s' % self.broker_url)
            if r.status_code == 200:
                payload = json.loads(r.text)
                if not self.broker_authid and not self.broker_password:
                    self.broker_noauth = True
                elif 'next-authid' in payload:
                    self.broker_authid = payload['next-authid']
                self.broker_username = broker_username or ''
                self.broker_password = broker_password or ''
                self._broker_auth_successful = True
                self.populate_session_profiles()
                return True
        self._broker_auth_successful = False
        self.broker_authid = None
        return False

    def broker_disconnect(self):
        """\
        Disconnect from an (already) authenticated broker session.

        All authentication parameters will be dropped (forgotten) and
        this instance has to re-authenticate against / re-connect to the
        session broker before any new interaction with the broker is possible.


        """
        _profile_ids = copy.deepcopy(self.profile_ids)

        # forget nearly everything...
        for profile_id in _profile_ids:
            self.init_profile_cache(profile_id)
            try: del self._profile_metatypes[profile_id]
            except KeyError: pass
            try: self._profiles_need_profile_id_renewal.remove(profile_id)
            except ValueError: pass
            try: del self._cached_profile_ids[profile_id]
            except KeyError: pass
            del self.session_profiles[profile_id]
        self._mutable_profile_ids = None
        self._broker_auth_successful = False
        self.broker_authid = None
        self.broker_password = None
        self.broker_noauth = False

    def is_broker_authenticated(self):
        """\
        Detect if an authenticated broker session has already been
        initiated. Todo so, a simple re-authentication (username, password)
        will be attempted. If that fails, user credentials are not provided /
        valid.


        :returns: ``True`` if the broker session has already been authenticated
            and user credentials are known / valid

        :rtype: ``bool``

        """
        if self._broker_auth_successful is None:
            # do a test auth against the given broker URL
            try:
                self.broker_simpleauth(self.broker_username, self.broker_password)
            except x2go.x2go_exceptions.X2GoBrokerConnectionException:
                self._broker_auth_successful = False
        return self._broker_auth_successful

    def broker_listprofiles(self):
        """\
        Obtain a session profile list from the X2Go Session Broker.


        :returns: session profiles as a Python dictionary.

        :rtype: ``dict``

        """
        if self.broker_url is not None:
            request_data = {
                'task': 'listprofiles',
                'user': self.broker_username,
            }
            if self.broker_authid is not None:
                request_data['authid'] = self.broker_authid
                self.logger("Sending request to broker: user: {user}, authid: {authid}, task: {task}".format(**request_data), log.loglevel_DEBUG)
            else:
                if self.broker_password:
                    request_data['password'] = "<hidden>"
                else:
                    request_data['password'] = "<EMPTY>"
                self.logger("Sending request to broker: user: {user}, password: {password}, task: {task}".format(**request_data), log.loglevel_DEBUG)
                request_data['password'] = self.broker_password or ''
            try:
                r = requests.post(self.broker_url, data=request_data)
            except requests.exceptions.ConnectionError:
                raise x2go.x2go_exceptions.X2GoBrokerConnectionException('Failed to connect to URL %s' % self.broker_url)
            if r.status_code == 200 and r.headers['content-type'].startswith("text/json"):
                payload = json.loads(r.text)
                if 'next-authid' in payload:
                    self.broker_authid = payload['next-authid']
                if 'mutable_profile_ids' in payload:
                    self._mutable_profile_ids = payload['mutable_profile_ids']
                self._broker_auth_successful = True
                return payload['profiles'] if payload['task'] == 'listprofiles' else {}
        self._broker_auth_successful = False
        self.broker_authid = None
        return {}

    def broker_selectsession(self, profile_id):
        """\
        Select a session from the list of available session profiles (presented by
        :func:`broker_listprofiles()`). This method requests a session information dictionary
        (server, port, SSH keys, already running / suspended sessions, etc.) from the
        session broker for the provided ``profile_id``.

        :param profile_id: profile ID of the selected session profile
        :type profile_id: ``str``
        :returns: session information (server, port, SSH keys, etc.) for a selected
            session profile (i.e. ``profile_id``)
        :rtype: ``dict``

        """
        if self.broker_url is not None:
            if profile_id not in self._broker_profile_cache or not self._broker_profile_cache[profile_id]:
                request_data = {
                    'task': 'selectsession',
                    'profile-id': profile_id,
                    'user': self.broker_username,
                    'pubkey': self.broker_my_pubkey,
                }
                if self.broker_authid is not None:
                    request_data['authid'] = self.broker_authid
                    self.logger("Sending request to broker: user: {user}, authid: {authid}, task: {task}".format(**request_data), log.loglevel_DEBUG)
                else:
                    if self.broker_password:
                        request_data['password'] = "<hidden>"
                    else:
                        request_data['password'] = "<EMPTY>"
                    self.logger("Sending request to broker: user: {user}, password: {password}, task: {task}".format(**request_data), log.loglevel_DEBUG)
                    request_data['password'] = self.broker_password or ''
                try:
                    r = requests.post(self.broker_url, data=request_data)
                except requests.exceptions.ConnectionError:
                    raise x2go.x2go_exceptions.X2GoBrokerConnectionException('Failed to connect to URL %s' % self.broker_url)
                if r.status_code == 200 and r.headers['content-type'].startswith("text/json"):
                    payload = json.loads(r.text)
                    if 'next-authid' in payload:
                        self.broker_authid = payload['next-authid']
                    self._broker_profile_cache[profile_id] = payload['selected_session'] if payload['task'] == 'selectsession' else {}
                    self._broker_auth_successful = True
                else:
                    self.broker_authid = None
                    self._broker_auth_successful = False
            self._broker_profile_cache[profile_id]
            return self._broker_profile_cache[profile_id]
        return {}

    def _init_profile_cache(self, profile_id):
        if str(profile_id) in self._broker_profile_cache:
            del self._broker_profile_cache[str(profile_id)]

    def _populate_session_profiles(self):
        """\
        Populate the set of session profiles by loading the session
        profile configuration from a file in INI format.


        :returns: a set of session profiles

        :rtype: ``dict``

        """
        if self.is_broker_authenticated() and \
           self.broker_noauth or \
           self.broker_username and self.broker_password:

            session_profiles = self.broker_listprofiles()
            _session_profiles = copy.deepcopy(session_profiles)

            for session_profile in _session_profiles:
                session_profile = str(session_profile)
                for key, default_value in list(self.defaultSessionProfile.items()):
                    key = str(key)
                    if type(default_value) is bytes:
                        default_value = str(default_value)
                    if key not in session_profiles[session_profile]:
                        session_profiles[session_profile][key] = default_value

        else:
            session_profiles = {}

        return session_profiles

    def _is_mutable(self, profile_id):
        if type(self._mutable_profile_ids) is list and profile_id in self._mutable_profile_ids:
            return True
        return False

    def _supports_mutable_profiles(self):
        if type(self._mutable_profile_ids) is list:
            return True
        return False

    def _write(self):
        print("not suported, yet")

    def _delete_profile(self, profile_id):
        del self.session_profiles[str(profile_id)]

    def _update_value(self, profile_id, option, value):
        if type(value) is bytes:
            value = str(value)
        self.session_profiles[str(profile_id)][str(option)] = value

    def _get_profile_parameter(self, profile_id, option, key_type):
        return key_type(self.session_profiles[str(profile_id)][str(option)])

    def _get_profile_options(self, profile_id):
        return list(self.session_profiles[str(profile_id)].keys())

    def _get_profile_ids(self):
        list(self.session_profiles.keys())
        return list(self.session_profiles.keys())

    def _get_server_hostname(self, profile_id):
        selected_session = self.broker_selectsession(profile_id)
        return selected_session['server']

    def _get_server_port(self, profile_id):
        selected_session = self.broker_selectsession(profile_id)
        return int(selected_session['port'])

    def _get_pkey_object(self, profile_id):
        selected_session = self.broker_selectsession(profile_id)
        if 'authentication_pubkey' in selected_session and selected_session['authentication_pubkey'] == 'ACCEPTED':
            time.sleep(2)
            return self.broker_my_privkey
        return None



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoClientSettings class - managing x2goclient settings file.

The :class:`x2go.backends.settings.file.X2GoClientSettings` class one of Python X2Go's a public API classes.
Use this class (indirectly by retrieving it from an :class:`x2go.client.X2GoClient` instance)
in your Python X2Go based applications to access the
»settings« configuration file of your X2Go client application.

This class supports reading the »settings« configuration from a file (``~/.x2goclient/settings``).

"""
__NAME__ = 'x2gosettings-pylib'

__package__ = 'x2go.backends.settings'
__name__    = 'x2go.backends.settings.file'

# Python X2Go modules
import x2go.log as log
from x2go.defaults import X2GO_SETTINGS_CONFIGFILES as _X2GO_SETTINGS_CONFIGFILES
from x2go.defaults import X2GO_CLIENTSETTINGS_DEFAULTS as _X2GO_CLIENTSETTINGS_DEFAULTS
import x2go.inifiles as inifiles


class X2GoClientSettings(inifiles.X2GoIniFile):
    """\
    Configuration file based settings for :class:`x2go.client.X2GoClient` instances.


    """
    def __init__(self, config_files=_X2GO_SETTINGS_CONFIGFILES, defaults=_X2GO_CLIENTSETTINGS_DEFAULTS, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        Constructs an :class:`x2go.backends.settings.file.X2GoClientSettings` instance. This is normally done from within an :class:`x2go.client.X2GoClient` instance.
        You can retrieve this :class:`x2go.backends.settings.file.X2GoClientSettings` instance with the :func:`X2GoClient.get_client_settings() <x2go.client.X2GoClient.get_client_settings()>`
        method.

        On construction the :class:`x2go.backends.settings.file.X2GoClientSettings` object is filled with values from the configuration files::

            /etc/x2goclient/settings
            ~/.x2goclient/settings

        The files are read in the specified order and config options of both files are merged. Options
        set in the user configuration file (``~/.x2goclient/settings``) override global options set in
        ``/etc/x2goclient/settings``.

        """
        inifiles.X2GoIniFile.__init__(self, config_files, defaults=defaults, logger=logger, loglevel=loglevel)



# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2Go reverse SSH/Paramiko tunneling provides X2Go sound, X2Go printing and
X2Go sshfs for folder sharing and mounting remote devices in X2Go terminal
server sessions.

"""
__NAME__ = 'x2gorevtunnel-pylib'

__package__ = 'x2go'
__name__    = 'x2go.rforward'

# modules
import copy
import threading
import gevent
import paramiko

# gevent/greenlet
from gevent import select, socket, Timeout

# Python X2Go modules
from . import log


def x2go_transport_tcp_handler(chan, origin, server):
    """\
    An X2Go customized TCP handler for the Paramiko/SSH ``Transport()`` class.

    Incoming channels will be put into Paramiko's default accept queue. This corresponds to
    the default behaviour of Paramiko's ``Transport`` class.

    However, additionally this handler function checks the server port of the incoming channel
    and detects if there are Paramiko/SSH reverse forwarding tunnels waiting for the incoming
    channels. The Paramiko/SSH reverse forwarding tunnels are initiated by an :class:`x2go.session.X2GoSession` instance
    (currently supported: reverse tunneling auf audio data, reverse tunneling of SSH requests).

    If the server port of an incoming Paramiko/SSH channel matches the configured port of an :class:`x2go.rforward.X2GoRevFwTunnel`
    instance, this instance gets notified of the incoming channel and a new :class:`x2go.rforward.X2GoRevFwChannelThread` is
    started. This :class:`x2go.rforward.X2GoRevFwChannelThread` then takes care of the new channel's incoming data stream.

    :param chan: a Paramiko channel object
    :type chan: ``paramiko.Channel`` object
    :param origin: host/port tuple where a connection originates from
    :type origin: ``tuple``
    :param server: host/port tuple where to connect to
    :type server: ``tuple``

    """
    (origin_addr, origin_port) = origin
    (server_addr, server_port) = server
    transport = chan.get_transport()
    transport._queue_incoming_channel(chan)
    rev_tuns = transport.reverse_tunnels

    for session_name in list(rev_tuns.keys()):

        if int(server_port) in [ int(tunnel[0]) for tunnel in list(rev_tuns[session_name].values()) ]:

            if rev_tuns[session_name]['snd'] is not None and int(server_port) == int(rev_tuns[session_name]['snd'][0]):
                rev_tuns[session_name]['snd'][1].notify()

            elif rev_tuns[session_name]['sshfs'] is not None and int(server_port) == int(rev_tuns[session_name]['sshfs'][0]):
                rev_tuns[session_name]['sshfs'][1].notify()


class X2GoRevFwTunnel(threading.Thread):
    """\
    :class:`x2go.rforward.X2GoRevFwTunnel` class objects are used to reversely tunnel
    X2Go audio, X2Go printing and X2Go folder sharing / device mounting
    through Paramiko/SSH.


    """
    def __init__(self, server_port, remote_host, remote_port, ssh_transport, session_instance=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        Setup a reverse tunnel through Paramiko/SSH.

        After the reverse tunnel has been setup up with :func:`X2GoRevFwTunnel.start() <x2go.rforward.X2GoRevFwTunnel.start()>` it waits
        for notification from :func:`X2GoRevFwTunnel.notify() <x2go.rforward.X2GoRevFwTunnel.notify()>` to accept incoming channels. This
        notification (:func:`X2GoRevFwTunnel.notify() <x2go.rforward.X2GoRevFwTunnel.notify()>` gets called from within the transport's
        TCP handler function :func:`x2go_transport_tcp_handler()` of the :class:`x2go.session.X2GoSession` instance.

        :param server_port: the TCP/IP port on the X2Go server (starting point of the tunnel),
            normally some number above 30000
        :type server_port: int
        :param remote_host: the target address for reversely tunneled traffic. With X2Go this should
            always be set to the localhost (IPv4) address.
        :type remote_host: str
        :param remote_port: the TCP/IP port on the X2Go client (end point of the tunnel),
            normally an application's standard port (22 for SSH, 4713 for pulse audio, etc.)
        :type remote_port: int
        :param ssh_transport: the :class:`x2go.session.X2GoSession`'s Paramiko/SSH transport instance
        :type ssh_transport: ``paramiko.Transport`` instance
        :param logger: you can pass an :class:`x2go.log.X2GoLogger` object to the
            :class:`x2go.rforward.X2GoRevFwTunnel` constructor
        :type logger: :class:`x2go.log.X2GoLogger` instance
        :param loglevel: if no :class:`x2go.log.X2GoLogger` object has been supplied a new one will be
            constructed with the given loglevel
        :type loglevel: int

        """
        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self.server_port = server_port
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.ssh_transport = ssh_transport
        self.session_instance = session_instance

        self.open_channels = {}
        self.incoming_channel = threading.Condition()

        threading.Thread.__init__(self)
        self.daemon = True
        self._accept_channels = True

    def __del__(self):
        """\
        Class destructor.

        """
        self.stop_thread()
        self.cancel_port_forward('', self.server_port)

    def cancel_port_forward(self, address, port):
        """\
        Cancel a port forwarding request. This cancellation request is sent to the server and
        on the server the port forwarding should be unregistered.

        :param address: remote server address
        :type address: ``str``
        :param port: remote port
        :type port: ``int``

        """
        timeout = Timeout(10)
        timeout.start()
        try:
            self.ssh_transport.global_request('cancel-tcpip-forward', (address, port), wait=True)
        except:
            pass
        finally:
            timeout.cancel()

    def pause(self):
        """\
        Prevent acceptance of new incoming connections through the Paramiko/SSH
        reverse forwarding tunnel. Also, any active connection on this :class:`x2go.rforward.X2GoRevFwTunnel`
        instance will be closed immediately, if this method is called.


        """
        if self._accept_channels == True:
            self.cancel_port_forward('', self.server_port)
            self._accept_channels = False
            self.logger('paused thread: %s' % repr(self), loglevel=log.loglevel_DEBUG)

    def resume(self):
        """\
        Resume operation of the Paramiko/SSH reverse forwarding tunnel
        and continue accepting new incoming connections.


        """
        if self._accept_channels == False:
            self._accept_channels = True
            self._requested_port = self.ssh_transport.request_port_forward('127.0.0.1', self.server_port, handler=x2go_transport_tcp_handler)
            self.logger('resumed thread: %s' % repr(self), loglevel=log.loglevel_DEBUG)

    def notify(self):
        """\
        Notify an :class:`x2go.rforward.X2GoRevFwTunnel` instance of an incoming Paramiko/SSH channel.

        If an incoming reverse tunnel channel appropriate for this instance has
        been detected, this method gets called from the :class:`x2go.session.X2GoSession`'s transport
        TCP handler.

        The sent notification will trigger a ``thread.Condition()`` waiting for notification
        in :func:`X2GoRevFwTunnel.run() <x2go.rforward.X2GoRevFwTunnel.run()>`.


        """
        self.incoming_channel.acquire()
        self.logger('notifying thread of incoming channel: %s' % repr(self), loglevel=log.loglevel_DEBUG)
        self.incoming_channel.notify()
        self.incoming_channel.release()

    def stop_thread(self):
        """\
        Stops this :class:`x2go.rforward.X2GoRevFwTunnel` thread completely.


        """
        self.pause()
        self._keepalive = False
        self.logger('stopping thread: %s' % repr(self), loglevel=log.loglevel_DEBUG)
        self.notify()

    def _request_port_forwarding(self):
        try:
            self._requested_port = self.ssh_transport.request_port_forward('127.0.0.1', self.server_port, handler=x2go_transport_tcp_handler)
        except paramiko.SSHException:
            # if port forward request fails, we try to tell the server to cancel all foregoing port forward requests on
            # self.server_port
            self.cancel_port_forward('', self.server_port)
            gevent.sleep(1)
            try:
                self._requested_port = self.ssh_transport.request_port_forward('127.0.0.1', self.server_port, handler=x2go_transport_tcp_handler)
            except paramiko.SSHException as e:
                if self.session_instance:
                    self.session_instance.HOOK_rforward_request_denied(server_port=self.server_port)
                else:
                    self.logger('Encountered SSHException: %s (for reverse TCP port forward with local destination port %s' % (str(e), self.server_port), loglevel=log.loglevel_WARN)

    def run(self):
        """\
        This method gets run once an :class:`x2go.rforward.X2GoRevFwTunnel` has been started with its
        :func:`start()` method. Use :class:`x2go.rforward.X2GoRevFwTunnel`.stop_thread() to stop the
        reverse forwarding tunnel again. You can also temporarily lock the tunnel
        down with :func:`X2GoRevFwTunnel.pause() <x2go.rforward.X2GoRevFwTunnel.pause()>` and :func:`X2GoRevFwTunnel.resume() <x2go.rforward.X2GoRevFwTunnel.resume()>`).

        :func:`X2GoRevFwTunnel.run() <x2go.rforward.X2GoRevFwTunnel.run()>` waits for notifications of an appropriate incoming
        Paramiko/SSH channel (issued by :func:`X2GoRevFwTunnel.notify() <x2go.rforward.X2GoRevFwTunnel.notify()>`). Appropriate in
        this context means, that its start point on the X2Go server matches the class's
        property ``server_port``.

        Once a new incoming channel gets announced by the :func:`notify()` method, a new
        :class:`x2go.rforward.X2GoRevFwChannelThread` instance will be initialized. As a data stream handler,
        the function :func:`x2go_rev_forward_channel_handler()` will be used.

        The channel will last till the connection gets dropped on the X2Go server side or
        until the tunnel gets paused by an :func:`X2GoRevFwTunnel.pause() <x2go.rforward.X2GoRevFwTunnel.pause()>` call or stopped via the
        :func:`X2GoRevFwTunnel.stop_thread() <x2go.rforward.X2GoRevFwTunnel.stop_thread()>` method.


        """
        self._request_port_forwarding()
        self._keepalive = True
        while self._keepalive:

            self.incoming_channel.acquire()

            self.logger('waiting for incoming data channel on X2Go server port: [127.0.0.1]:%s' % self.server_port, loglevel=log.loglevel_DEBUG)
            self.incoming_channel.wait()

            if self._keepalive:
                self.logger('detected incoming data channel on X2Go server port: [127.0.0.1]:%s' % self.server_port, loglevel=log.loglevel_DEBUG)
                _chan = self.ssh_transport.accept()
                self.logger('data channel %s for server port [127.0.0.1]:%s is up' % (_chan, self.server_port), loglevel=log.loglevel_DEBUG)
            else:
                self.logger('closing down rev forwarding tunnel on remote end [127.0.0.1]:%s' % self.server_port, loglevel=log.loglevel_DEBUG)

            self.incoming_channel.release()
            if self._accept_channels and self._keepalive:
                _new_chan_thread = X2GoRevFwChannelThread(_chan, (self.remote_host, self.remote_port),
                                                          target=x2go_rev_forward_channel_handler,
                                                          kwargs={
                                                            'chan': _chan,
                                                            'addr': self.remote_host,
                                                            'port': self.remote_port,
                                                            'parent_thread': self,
                                                            'logger': self.logger,
                                                          }
                                                  )
                _new_chan_thread.start()
                self.open_channels['[%s]:%s' % _chan.origin_addr] = _new_chan_thread


def x2go_rev_forward_channel_handler(chan=None, addr='', port=0, parent_thread=None, logger=None, ):
    """\
    Handle the data stream of a requested channel that got set up by a :class:`x2go.rforward.X2GoRevFwTunnel` (Paramiko/SSH
    reverse forwarding tunnel).

    The channel (and the corresponding connections) close either ...

        - ... if the connecting application closes the connection and thus, drops
          the channel, or
        - ... if the :class:`x2go.rforward.X2GoRevFwTunnel` parent thread gets paused. The call
          of :func:`X2GoRevFwTunnel.pause() <x2go.rforward.X2GoRevFwTunnel.pause()>` on the instance can be used to shut down all incoming
          tunneled SSH connections associated to this :class:`x2go.rforward.X2GoRevFwTunnel` instance
          from within a Python X2Go application.

    :param chan: channel (Default value = None)
    :type chan: ``class``
    :param addr: bind address (Default value = '')
    :type addr: ``str``
    :param port: bind port (Default value = 0)
    :type port: ``int``
    :param parent_thread: the calling :class:`x2go.rforward.X2GoRevFwTunnel` instance (Default value = None)
    :type parent_thread: :class:`x2go.rforward.X2GoRevFwTunnel` instance
    :param logger: you can pass an :class:`x2go.log.X2GoLogger` object to the
        :class:`x2go.rforward.X2GoRevFwTunnel` constructor (Default value = None)
    :type logger: :class:`x2go.log.X2GoLogger` instance

    """
    fw_socket = socket.socket()
    fw_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    if logger is None:
        def _dummy_logger(msg, l):
            pass
        logger = _dummy_logger

    try:
        fw_socket.connect((addr, port))
    except Exception as e:
        logger('Reverse forwarding request to %s:%d failed: %r' % (addr, port, e), loglevel=log.loglevel_INFO)
        return

    logger('Connected! Reverse tunnel open %r -> %r -> %r' % (chan.origin_addr,
                                                              chan.getpeername(), (addr, port)),
                                                              loglevel=log.loglevel_INFO)
    n = 0
    while parent_thread._accept_channels:
        r, w, x = select.select([fw_socket, chan], [], [])
        try:
            if fw_socket in r:
                data = fw_socket.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                fw_socket.send(data)
        except socket.error as e:
            logger('Reverse tunnel %s encoutered socket error: %s' % (chan, str(e)), loglevel=log.loglevel_WARN)
            n += 1
            if n >= 1024:
                break

    chan.close()
    fw_socket.close()
    logger('Reverse tunnel %s closed from %r' % (chan, chan.origin_addr,), loglevel=log.loglevel_INFO)


class X2GoRevFwChannelThread(threading.Thread):
    """\
    Starts a thread for each incoming Paramiko/SSH data channel trough the reverse
    forwarding tunnel.


    """
    def __init__(self, channel, remote=None, **kwargs):
        """\
        Initializes a reverse forwarding channel thread.

        :param channel: incoming Paramiko/SSH channel from the :class:`x2go.session.X2GoSession`'s transport
            accept queue
        :type channel: class
        :param remote: tuple (addr, port) that specifies the data endpoint of the channel
        :type remote: ``tuple(str, int)``

        """
        self.channel = channel
        if remote is not None:
            self.remote_host = remote[0]
            self.remote_port = remote[1]
        threading.Thread.__init__(self, **kwargs)
        self.daemon = True



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoProxy class - proxying your graphical connection through NX3.

"""
__NAME__ = 'x2goproxynx3-pylib'

__package__ = 'x2go.backends.proxy'
__name__    = 'x2go.backends.proxy.nx3'

# modules
import os

# Python X2Go modules
import x2go.log as log
import x2go.backends.proxy.base as base

from x2go.defaults import X2GOCLIENT_OS as _X2GOCLIENT_OS

class X2GoProxy(base.X2GoProxy):
    """\
    This :class:`x2go.backends.proxy.nx3.X2GoProxy` class is a NX version 3 based X2Go proxy connection class.

    It basically fills :class:`x2go.backends.proxy.base.X2GoProxy` variables with sensible content. Its
    methods mostly wrap around the corresponding methods of the parent class.


    """
    def __init__(self, *args, **kwargs):
        """\
        For available parameters refer to :class:`x2go.backends.proxy.base.X2GoProxy` class documentation.

        """
        base.X2GoProxy.__init__(self, *args, **kwargs)
        self.subsystem = 'NX Proxy'

        # setting some default environment variables, nxproxy paths etc.
        if _X2GOCLIENT_OS == "Windows":
            _nxproxy_paths = [
                os.path.join(os.environ["ProgramFiles"], os.path.normpath("PyHoca-GUI/nxproxy/nxproxy.exe")),
                os.path.join(os.environ["ProgramFiles"], os.path.normpath("x2goclient/nxproxy.exe")),
                os.path.join(os.environ["ProgramFiles"], os.path.normpath("NX Client for Windows/bin/nxproxy.exe")),
                os.path.normpath("../pyhoca-contrib/mswin/nxproxy-mswin/nxproxy-3.5.0.27_cygwin-2015-10-18/nxproxy.exe"),
            ]
            if 'NXPROXY_BINARY' in os.environ:
                _nxproxy_paths.insert(0, os.environ['NXPROXY_BINARY'])
            for _nxproxy_cmd in _nxproxy_paths:
                if os.path.exists(_nxproxy_cmd):
                    break
            self.PROXY_CMD = _nxproxy_cmd
        elif 'NXPROXY_BINARY' in os.environ:
            self.PROXY_CMD = os.environ['NXPROXY_BINARY']
        else:
            self.PROXY_CMD = "/usr/bin/nxproxy"
        self.PROXY_ENV.update({
            "NX_CLIENT": "/bin/true",
            "NX_ROOT": self.sessions_rootdir
        })
        self.PROXY_MODE = '-S'
        if _X2GOCLIENT_OS == "Windows":
            self.PROXY_OPTIONS = [
                "nx/nx" ,
                "retry=5",
                "composite=1",
                "connect=127.0.0.1",
                "clipboard=1",
                "cookie=%s" % self.session_info.cookie,
                "port=%s" % self.session_info.graphics_port,
                "errors=%s" % os.path.join(".", "..", "S-%s" % self.session_info.name, self.session_errors, ),
            ]
        else:
            self.PROXY_OPTIONS = [
                "nx/nx" ,
                "retry=5",
                "composite=1",
                "connect=127.0.0.1",
                "clipboard=1",
                "cookie=%s" % self.session_info.cookie,
                "port=%s" % self.session_info.graphics_port,
                "errors=%s" % os.path.join(self.session_info.local_container, self.session_errors, ),
            ]

        self.PROXY_DISPLAY = self.session_info.display

    def _update_local_proxy_socket(self, port):
        """\
        Update the local proxy socket on port changes due to already-bound-to local TCP/IP port sockets.

        :param port: new local TCP/IP socket port
        :type port: ``int``

        """

        for idx, a in enumerate(self.PROXY_OPTIONS):
            if a.startswith('port='):
                self.PROXY_OPTIONS[idx] = 'port=%s' % port

    def _generate_cmdline(self):
        """\
        Generate the NX proxy command line for execution.


        """
        if _X2GOCLIENT_OS == "Windows":
            _options_filename = os.path.join(self.session_info.local_container, 'options')
            options = open(_options_filename, 'w')
            options.write(u'%s:%s' % (','.join(self.PROXY_OPTIONS), self.PROXY_DISPLAY))
            options.close()
            self.PROXY_OPTIONS= [ 'nx/nx', 'options=%s' % os.path.join(".", "..", "S-%s" % self.session_info.name, 'options'), ]

        cmd_line = [ self.PROXY_CMD, ]
        cmd_line.append(self.PROXY_MODE)
        _proxy_options = "%s:%s" % (",".join(self.PROXY_OPTIONS), self.PROXY_DISPLAY)
        cmd_line.append(_proxy_options)
        return cmd_line

    def process_proxy_options(self):
        base.X2GoProxy.process_proxy_options(self)

    def start_proxy(self):
        """\
        Start the thread runner and wait for the proxy to come up.


        :returns: a subprocess instance that knows about the externally started proxy command.

        :rtype: ``obj``

        """
        self.logger('starting local NX3 proxy...', loglevel=log.loglevel_INFO)
        self.logger('NX3 Proxy mode is server, cookie=%s, host=127.0.0.1, port=%s.' % (self.session_info.cookie, self.session_info.graphics_port,), loglevel=log.loglevel_DEBUG)
        self.logger('NX3 proxy writes session log to %s.' % os.path.join(self.session_info.local_container, 'session.log'), loglevel=log.loglevel_DEBUG)

        p, p_ok = base.X2GoProxy.start_proxy(self)

        if self.ok():
            self.logger('NX3 proxy is up and running.', loglevel=log.loglevel_INFO)
        else:
            self.logger('Bringing up NX3 proxy failed.', loglevel=log.loglevel_ERROR)

        return p, self.ok()



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

__NAME__    = 'python-x2go'
__VERSION__ = '0.6.1.4'
__AUTHOR__  = 'Mike Gabriel'

__package__ = 'x2go'
__name__    = 'x2go'

import os
from .defaults import X2GOCLIENT_OS

if X2GOCLIENT_OS != 'Windows':
    # in Debian unstable, the ares resolver is broken, see
    # https://bugs.debian.org/908144
    #os.environ["GEVENT_RESOLVER"] = "ares"

    # we could use the dnspython resolver...
    #os.environ["GEVENT_RESOLVER"] = "dnspython"

    # however, this is unsupported in Debian 9, so we do nothing here...
    pass

from gevent import monkey
monkey.patch_all()

from . import utils

from .client import X2GoClient
from .backends.profiles.file import X2GoSessionProfiles
from .backends.printing.file import X2GoClientPrinting
from .backends.settings.file import X2GoClientSettings
from .session import X2GoSession
from .sshproxy import X2GoSSHProxy
from .x2go_exceptions import *
from .log import *

from .cleanup import x2go_cleanup

from .defaults import CURRENT_LOCAL_USER
from .defaults import LOCAL_HOME
from .defaults import X2GO_CLIENT_ROOTDIR
from .defaults import X2GO_SESSIONS_ROOTDIR
from .defaults import X2GO_SSH_ROOTDIR
from .defaults import BACKENDS

if X2GOCLIENT_OS == 'Windows':
    from .xserver import X2GoClientXConfig, X2GoXServer

# compat section
X2goClient = X2GoClient
X2goSessionProfiles = X2GoSessionProfiles
X2goClientPrinting = X2GoClientPrinting
X2goClientSettings = X2GoClientSettings
X2goSession = X2GoSession
X2goSSHProxy = X2GoSSHProxy

if X2GOCLIENT_OS == 'Windows':
    X2goClientXConfig = X2GoClientXConfig
    X2goXServer = X2GoXServer



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
A recommended X2Go session clean up helper function.

"""

__package__ = 'x2go'
__name__    = 'x2go.cleanup'

import gevent
import paramiko
import threading

# Python X2Go modules
from . import rforward
from .defaults import X2GOCLIENT_OS as _X2GOCLIENT_OS

if _X2GOCLIENT_OS == 'Windows':
    from . import xserver
    from . import pulseaudio

def x2go_cleanup(e=None, threads=None):
    """\
    For every Python X2Go application you write, please make sure to
    capture the ``KeyboardInterrupt`` and the ``SystemExit`` exceptions and
    call this function if either of the exceptions occurs.

    Example::

        import x2go

        try:
            my_x2goclient = x2go.X2GoClient(...)

            [... your code ...]

            sys.exit(0)
        except (KeyboardInterrupt, SystemExit):
            x2go.x2go_cleanup()

    :param e: if :func:`x2go_cleanup()` got called as you caught an exception in your code this can be the
        ``Exception`` that we will process at the end of the clean-up (or if clean-up failed or was not
        appropriate) (Default value = None)
    :type e: ``exception``
    :param threads: a list of threads to clean up (Default value = None)
    :type threads: ``list``

    """
    try:
        if threads is None:
            threads = threading.enumerate()
        else:
            threads = threads

        # stop X2Go reverse forwarding tunnels
        for t in threads:
            if type(t) == rforward.X2GoRevFwTunnel:
                t.stop_thread()
                del t

        # stop X2Go paramiko transports used by X2GoTerminalSession objects
        for t in threads:
            if type(t) == paramiko.Transport:
                if hasattr(t, '_x2go_session_marker'):
                    t.stop_thread()
                    del t

        # on Windows: stop the XServer that we evoked
        if _X2GOCLIENT_OS == 'Windows':
            for t in threads:
                if type(t) == xserver.X2GoXServer:
                    t.stop_thread()
                    del t

        # on Windows: stop the PulseAudio daemon that we evoked
        if _X2GOCLIENT_OS == 'Windows':
            for t in threads:
                if type(t) == pulseaudio.X2GoPulseAudio:
                    t.stop_thread()
                    del t

        for t in threads:
            # now let's catch X2GoSessionGuardian, which is a bit tricky
            # as we need to avoid circular imports
            if hasattr(t, 'stop_thread') and hasattr(t, 'guardian'):
                t.stop_thread()
                del t

        gevent.sleep(1)

        if e is not None:
            raise e

    except KeyboardInterrupt:
        # do not allow keyboard interrupts during Python X2Go cleanup
        pass
    except SystemExit:
        # neither do we allow SIGTERM signals during cleanup...
        pass



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
:class:`x2go.backends.profiles.sshbroker.X2GoSessionProfiles` class - managing X2Go Client  session profiles obtained from an SSH based X2Go Session Broker.

:class:`x2go.backends.profiles.sshbroker.X2GoSessionProfiles` is a public API class. Use this class in your Python X2Go based
applications.

"""
__NAME__ = 'x2gosessionprofiles-pylib'

__package__ = 'x2go.backends.profiles'
__name__ = 'x2go.backends.profiles.sshbroker'

# modules
import copy

# Python X2Go modules
import x2go.backends.profiles.base as base
import x2go.log as log

from x2go.defaults import X2GO_SESSIONPROFILE_DEFAULTS as _X2GO_SESSIONPROFILE_DEFAULTS

from x2go.x2go_exceptions import X2GoNotImplementedYetException

class X2GoSessionProfiles(base.X2GoSessionProfiles):

    defaultSessionProfile = copy.deepcopy(_X2GO_SESSIONPROFILE_DEFAULTS)
    _non_profile_sections = ('embedded')

    def __init__(self, session_profile_defaults=_X2GO_SESSIONPROFILE_DEFAULTS, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        Retrieve X2Go session profiles from a SSH session broker.

        :param session_profile_defaults: a default session profile
        :type session_profile_defaults: ``dict``
        :param logger: you can pass an :class:`x2go.log.X2GoLogger` object to the
            :class:`x2go.backends.profiles.httpbroker.X2GoSessionProfiles` constructor
        :type logger: :class:`x2go.log.X2GoLogger` instance
        :param loglevel: if no :class:`x2go.log.X2GoLogger` object has been supplied a new one will be
            constructed with the given loglevel
        :type loglevel: ``int``

        """
        raise X2GoNotImplementedYetException('HTTPSBROKER backend support is not implemented yet')



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Other contributors:
#       none so far

"""\
X2GoPulseAudio class - a Pulseaudio daemon guardian thread.

"""

__NAME__ = 'x2gopulseaudio-pylib'

__package__ = 'x2go'
__name__    = 'x2go.pulseaudio'

from .defaults import X2GOCLIENT_OS as _X2GOCLIENT_OS
if _X2GOCLIENT_OS == 'Windows':
    import win32process
    import win32con
    import win32event

# modules
import os
import sys
import threading
import gevent
import copy
import socket

from .defaults import LOCAL_HOME as _LOCAL_HOME

# Python X2Go modules
from . import log

class OSNotSupportedException(BaseException): pass
""" Exception denoting that this operating system is not supported. """

class X2GoPulseAudio(threading.Thread):
    """This class controls the Pulse Audio daemon."""

    def __init__(self, path=None, client_instance=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        Initialize a Pulse Audio daemon instance.

        :param path: full path to pulseaudio.exe
        :type path: ``str``
        :param client_instance: the calling :class:`x2go.client.X2GoClient` instance
        :type client_instance: :class:`x2go.client.X2GoClient` instance
        :param logger: you can pass an :class:`x2go.log.X2GoLogger` object to the :class:`x2go.xserver.X2GoClientXConfig` constructor
        :type logger: ``obj``
        :param loglevel: if no :class:`x2go.log.X2GoLogger` object has been supplied a new one will be
            constructed with the given loglevel
        :type loglevel: ``int``

        :raises OSNotSupportedException: on non-Windows platforms Python X2Go presumes that pulseaudio is already launched

        """
        if _X2GOCLIENT_OS not in ("Windows"):
            raise OSNotSupportedException('classes of x2go.pulseaudio module are for Windows only')

        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self.path = path
        self.client_instance = client_instance
        self._keepalive = None

        threading.Thread.__init__(self)
        self.daemon = True
        self.start()

    def run(self):
        """\
        This method is called once the ``X2GoPulseAudio.start()`` method has been called. To tear
        down the Pulseaudio daemon call the :func:`X2GoPulseAudio.stop_thread() <x2go.pulseaudio.X2GoPulseAudio.stop_thread()` method.


        """
        self._keepalive = True
        cmd = 'pulseaudio.exe'
        cmd_options = [
            '-n',
            '--exit-idle-time=-1',
            '-L "module-native-protocol-tcp port=4713 auth-cookie=\\\\.pulse-cookie"',
            '-L module-waveout',
        ]

        # Fix for python-x2go bug #537.
        # Works Around PulseAudio bug #80772.
        # Tested with PulseAudio 5.0.
        # This argument will not cause PulseAudio 1.1 to fail to launch.
        # However, PulseAudio 1.1 ignores it for some reason.
        # So yes, the fact that 1.1 ignores it would be a bug in python-x2go if
        # we ever ship 1.1 again.
        #
        # wv.major == 5 matches 2000, XP, and Server 2003 (R2).
        # (Not that we support 2000.)
        wv = sys.getwindowsversion()
        if (wv.major==5):
            self.logger('Windows XP or Server 2003 (R2) detected.', loglevel=log.loglevel_DEBUG)
            self.logger('Setting PulseAudio to "Normal" CPU priority.', loglevel=log.loglevel_DEBUG)
            cmd_options.insert(0,"--high-priority=no")

        cmd_options = " %s" % " ".join(cmd_options)

        if not os.path.isdir(os.path.join(_LOCAL_HOME, '.pulse', '%s-runtime' % socket.gethostname())):
            os.makedirs(os.path.join(_LOCAL_HOME, '.pulse', '%s-runtime' % socket.gethostname()))
        self.logger('starting PulseAudio server with command line: %s%s' % (cmd, cmd_options), loglevel=log.loglevel_DEBUG)

        si = win32process.STARTUPINFO()
        p_info = win32process.CreateProcess(None,
                                            '%s\\%s %s' % (self.path, cmd, cmd_options),
                                            None,
                                            None,
                                            0,
                                            win32con.CREATE_NO_WINDOW|win32process.NORMAL_PRIORITY_CLASS,
                                            None,
                                            None,
                                            si,
                                           )
        (hProcess, hThread, processId, threadId) = p_info

        gevent.sleep(5)
        rc = win32event.WaitForMultipleObjects([hProcess],
                                               1,
                                               1, # wait just one millisec
                                              )
        _is_alive = ( rc != win32event.WAIT_OBJECT_0 )
        if self.client_instance and not _is_alive:
            if 'CLIENTNAME' in os.environ:
                self.client_instance.HOOK_pulseaudio_not_supported_in_RDPsession()
            else:
                self.client_instance.HOOK_pulseaudio_server_startup_failed()

        while self._keepalive and _is_alive:
            gevent.sleep(1)
            rc = win32event.WaitForMultipleObjects([hProcess],
                                               1,
                                               1, # wait just one millisec
                                              )
            _is_alive = ( rc != win32event.WAIT_OBJECT_0 )
            if self.client_instance and not _is_alive:
                self.client_instance.HOOK_pulseaudio_server_died()

        self.logger('terminating running PulseAudio server', loglevel=log.loglevel_DEBUG)

        # there is no real kill command on Windows...
        self.logger('PulseAudio process ID to terminate: %s' % processId, loglevel=log.loglevel_DEBUG)
        try:
            win32process.TerminateProcess(hProcess, 0)
        except win32process.error:
            pass

    def stop_thread(self):
        """\
        Tear down a running Pulseaudio daemon.


        """
        self.logger('stop_thread() method has been called', loglevel=log.loglevel_DEBUG)
        self._keepalive = False




# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoSessionGuardian class - a guardian thread that controls X2Go session threads
and their sub-threads (like reverse forwarding tunnels, Paramiko transport threads,
etc.).

"""
__NAME__ = 'x2goguardian-pylib'

__package__ = 'x2go'
__name__    = 'x2go.guardian'

# modules
import gevent
import threading
import copy

# Python X2Go modules
from .cleanup import x2go_cleanup
from . import log

class X2GoSessionGuardian(threading.Thread):
    """\
    :class:`x2go.guardian.X2GoSessionGuardian` thread controls X2Go session threads and their sub-threads (like
    reverse forwarding tunnels, Paramiko transport threads, etc.). Its main function is
    to tidy up once a session gets interrupted (SIGTERM, SIGINT).

    There is one :class:`x2go.guardian.X2GoSessionGuardian` for each :class:`x2go.client.X2GoClient` instance (thus: for normal
    setups there should be _one_ :class:`x2go.client.X2GoClient` and _one_ :class:`x2go.guardian.X2GoSessionGuardian` in use).


    """
    def __init__(self, client_instance,
                 auto_update_listsessions_cache=False,
                 auto_update_listdesktops_cache=False,
                 auto_update_listmounts_cache=False,
                 auto_update_sessionregistry=False,
                 auto_register_sessions=False,
                 no_auto_reg_pubapp_sessions=False,
                 refresh_interval=5,
                 logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        :param auto_update_listsessions_cache: let :class:`x2go.guardian.X2GoSessionGuardian` refresh the session list cache for all :class:`x2go.session.X2GoSession` objects
        :type auto_update_listsessions_cache: ``bool``
        :param auto_update_listdesktops_cache: let :class:`x2go.guardian.X2GoSessionGuardian` refresh desktop lists in the session list cache for all :class:`x2go.session.X2GoSession` objects
        :type auto_update_listdesktops_cache: ``bool``
        :param auto_update_listmounts_cache: let :class:`x2go.guardian.X2GoSessionGuardian` refresh mount lists in the session list cache for all :class:`x2go.session.X2GoSession` objects
        :type auto_update_listmounts_cache: ``bool``
        :param auto_update_sessionregistry: if set to ``True`` the session status will be updated in regular intervals
        :type auto_update_sessionregistry: ``bool``
        :param auto_register_sessions: register new sessions automatically once they appear in the X2Go session (e.g.
            instantiated by another client that is connected to the same X2Go server under same user ID)
        :type auto_register_sessions: ``bool``
        :param no_auto_reg_pubapp_sessions: do not auto-register published applications sessions
        :type no_auto_reg_pubapp_sessions: ``bool``
        :param refresh_interval: refresh cache and session registry every <refresh_interval> seconds
        :type refresh_interval: ``int``
        :param logger: you can pass an :class:`x2go.log.X2GoLogger` object to the :class:`x2go.guardian.X2GoSessionGuardian` constructor
        :type logger: ``obj``
        :param loglevel: if no :class:`x2go.log.X2GoLogger` object has been supplied a new one will be
            constructed with the given loglevel
        :type loglevel: ``int``

        """
        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self.client_instance = client_instance
        self.auto_update_listsessions_cache = auto_update_listsessions_cache
        self.auto_update_listdesktops_cache = auto_update_listdesktops_cache
        self.auto_update_listmounts_cache = auto_update_listmounts_cache
        self.auto_update_sessionregistry = auto_update_sessionregistry
        self.auto_register_sessions = auto_register_sessions
        self.no_auto_reg_pubapp_sessions = no_auto_reg_pubapp_sessions
        self.refresh_interval = refresh_interval

        threading.Thread.__init__(self, target=self.guardian)
        self.daemon = True
        self.start()

    def guardian(self):
        """\
        The handler of this :class:`x2go.guardian.X2GoSessionGuardian` thread.


        """
        seconds = 0
        self._keepalive = True
        while self._keepalive:
            gevent.sleep(1)
            seconds += 1

            if seconds % self.refresh_interval == 0:

                self.logger('Entering X2Go Guardian client management loop...', loglevel=log.loglevel_DEBUG)

                if self.auto_update_listsessions_cache:
                    self.client_instance.update_cache_all_profiles(update_sessions=self.auto_update_listsessions_cache,
                                                                   update_desktops=self.auto_update_listdesktops_cache,
                                                                   update_mounts=self.auto_update_listmounts_cache,
                                                                  )

                if self.auto_update_sessionregistry and not self.auto_register_sessions:
                    self.client_instance.update_sessionregistry_status_all_profiles()

                # session auto-registration will automatically trigger an update of the session registry status
                if self.auto_register_sessions:
                    self.client_instance.register_available_server_sessions_all_profiles(skip_pubapp_sessions=self.no_auto_reg_pubapp_sessions)

        self.logger('X2Go session guardian thread waking up after %s seconds' % seconds, loglevel=log.loglevel_DEBUG)

        for session_uuid in list(self.client_instance.session_registry.keys()):
            session_summary = self.client_instance.get_session_summary(session_uuid)
            self.logger('calling session cleanup on profile %s for terminal session: %s' % (session_summary['profile_name'], session_summary['session_name']), loglevel=log.loglevel_DEBUG)
            x2go_cleanup(threads=session_summary['active_threads'])

    def stop_thread(self):
        """\
        Stop this :class:`x2go.guardian.X2GoSessionGuardian` thread.


        """
        self._keepalive = False



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoListSessionCache class - caching X2Go session information.

"""
__NAME__ = 'x2gocache-pylib'

__package__ = 'x2go'
__name__    = 'x2go.cache'

# modules
import copy
import gevent

# Python X2Go modules
from . import log
from . import x2go_exceptions

class X2GoListSessionsCache(object):
    """\
    For non-blocking operations in client applications using Python X2Go, it is
    recommended to enable the :class:`x2go.cache.X2GoListSessionsCache`. This can be done by calling
    the constructor of the :class:`x2go.client.X2GoClient` class.

    The session list and desktop cache gets updated in regular intervals by a threaded
    :class:`x2go.guardian.X2GoSessionGuardian` instance. For the session list and desktop list update, the
    X2Go server commands ``x2golistsessions`` and ``x2godesktopsessions`` are called and
    the command's stdout is cached in the session list cache.

    Whenever your client application needs access to either the server's session list
    or the server's desktop list the session cache is queried instead. This assures that
    the server's session/desktop list is available without delay, even on slow internet
    connections.


    """
    x2go_listsessions_cache = {}

    def __init__(self, client_instance, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        :param client_instance: the :class:`x2go.client.X2GoClient` instance that uses this :class:`x2go.cache.X2GoListSessionsCache`
        :type client_instance: ``obj``
        :param logger: you can pass an :class:`x2go.log.X2GoLogger` object to the :class:`x2go.cache.X2GoListSessionsCache` constructor
        :type logger: ``obj``
        :param loglevel: if no :class:`x2go.log.X2GoLogger` object has been supplied a new one will be
            constructed with the given loglevel
        :type loglevel: ``int``

        """
        self.x2go_listsessions_cache = {}
        self.last_listsessions_cache = {}
        self.protected = False

        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self.client_instance = client_instance

    def delete(self, profile_name):
        """\
        Remove session list from cache for a given profile.

        :param profile_name: name of profile to operate on
        :type profile_name: ``str``

        """
        while self.protected:
            gevent.sleep(.1)
        try: del self.x2go_listsessions_cache[profile_name]
        except KeyError: pass

    def check_cache(self):
        """\
        Check if session list cache elements are still valid (i.e. if all corresponding
        session profiles are still connected). If not so, remove invalid cache entries from
        the session list cache.


        """
        for profile_name in list(self.x2go_listsessions_cache.keys()):
            if profile_name not in self.client_instance.client_connected_profiles(return_profile_names=True):
                del self.x2go_listsessions_cache[profile_name]

    def update_all(self, update_sessions=True, update_desktops=False):
        """\
        Update :class:`x2go.cache.X2GoListSessionsCache` for all connected session profiles.

        :param update_sessions: cache recent session lists from all connected servers (Default value = True)
        :type update_sessions: ``bool``
        :param update_desktops: cache recent desktop lists from all connected servers (Default value = False)
        :type update_desktops: ``bool``

        """
        for profile_name in self.client_instance.client_connected_profiles(return_profile_names=True):
            self.update(profile_name, update_sessions=update_sessions, update_desktops=update_desktops)

        self.check_cache()

    def update(self, profile_name, update_sessions=True, update_desktops=False, update_mounts=False):
        """\
        Update :class:`x2go.cache.X2GoListSessionsCache` (i.e. session/desktops) for session profile ``profile_name``.

        :param profile_name: name of profile to update
        :type profile_name: ``str``
        :param update_sessions: cache recent session list from server (Default value = True)
        :type update_sessions: ``bool``
        :param update_desktops: cache recent desktop list from server (Default value = False)
        :type update_desktops: ``bool``
        :param update_mounts: cache list of client-side mounts on server (Default value = False)
        :type update_mounts: ``bool``

        """
        self.protected = True
        self.last_listsessions_cache = copy.deepcopy(self.x2go_listsessions_cache)
        control_session = self.client_instance.client_control_session_of_profile_name(profile_name)
        if profile_name not in self.x2go_listsessions_cache:
            self.x2go_listsessions_cache[profile_name] = {'sessions': None, 'desktops': None, 'mounts': {}, }
        if update_sessions:
            self._update_sessions(profile_name, control_session)
        if update_desktops:
            self._update_desktops(profile_name, control_session)
        if update_mounts:
            self._update_mounts(profile_name, control_session)
        self.protected = False

    def _update_mounts(self, profile_name, control_session):
        """\
        Update mounts list of :class:`x2go.cache.X2GoListSessionsCache` for session profile ``profile_name``.

        :param profile_name: name of profile to update
        :type profile_name: ``str``
        :param control_session: X2Go control session instance
        :raises X2GoControlSessionException: if the control session's ``list_mounts`` method fails

        """
        try:
            self.x2go_listsessions_cache[profile_name]['mounts'] = {}
            if self.x2go_listsessions_cache[profile_name]['sessions']:
                for session_name in self.x2go_listsessions_cache[profile_name]['sessions']:
                    session = self.client_instance.get_session_of_session_name(session_name, return_object=True, match_profile_name=profile_name)
                    if session is not None and session.is_running():
                        if control_session is not None and not control_session.has_session_died():
                            self.x2go_listsessions_cache[profile_name]['mounts'].update(control_session.list_mounts(session_name))
        except (x2go_exceptions.X2GoControlSessionException, AttributeError) as e:
            if profile_name in list(self.x2go_listsessions_cache.keys()):
                del self.x2go_listsessions_cache[profile_name]
            self.protected = False
            raise x2go_exceptions.X2GoControlSessionException(str(e))
        except x2go_exceptions.X2GoTimeOutException:
            pass
        except KeyError:
            pass

    def _update_desktops(self, profile_name, control_session):
        """\
        Update session lists of :class:`x2go.cache.X2GoListSessionsCache` for session profile ``profile_name``.

        :param profile_name: name of profile to update
        :type profile_name: ``str``
        :param control_session: X2Go control session instance
        :type control_session: ``obj``
        :raises X2GoControlSessionException: if the control session's ``list_desktop`` method fails

        """
        try:
            if control_session is not None and not control_session.has_session_died():
                self.x2go_listsessions_cache[profile_name]['desktops'] = control_session.list_desktops()
        except (x2go_exceptions.X2GoControlSessionException, AttributeError) as e:
            if profile_name in list(self.x2go_listsessions_cache.keys()):
                del self.x2go_listsessions_cache[profile_name]
            self.protected = False
            raise x2go_exceptions.X2GoControlSessionException(str(e))
        except x2go_exceptions.X2GoTimeOutException:
            pass
        except KeyError:
            pass

    def _update_sessions(self, profile_name, control_session):
        """\
        Update desktop list of :class:`x2go.cache.X2GoListSessionsCache` for session profile ``profile_name``.

        :param profile_name: name of profile to update
        :type profile_name: ``str``
        :param control_session: X2Go control session instance
        :type control_session: ``obj``
        :raises X2GoControlSessionException: if the control session's ``list_sessions`` method fails

        """
        try:
            if control_session is not None and not control_session.has_session_died():
                self.x2go_listsessions_cache[profile_name]['sessions'] = control_session.list_sessions()
        except (x2go_exceptions.X2GoControlSessionException, AttributeError) as e:
            if profile_name in list(self.x2go_listsessions_cache.keys()):
                del self.x2go_listsessions_cache[profile_name]
            self.protected = False
            raise x2go_exceptions.X2GoControlSessionException(str(e))
        except KeyError:
            pass

    def list_sessions(self, session_uuid):
        """\
        Retrieve a session list from the current cache content of :class:`x2go.cache.X2GoListSessionsCache`
        for a given :class:`x2go.session.X2GoSession` instance (specified by its unique session UUID).

        :param session_uuid: unique identifier of session to query cache for
        :type session_uuid: ``str``
        :returns: a data object containing available session information
        :rtype: ``X2GoServerSessionList*`` instance (or ``None``)

        """
        profile_name = self.client_instance.get_session_profile_name(session_uuid)
        if self.is_cached(session_uuid=session_uuid):
            return self.x2go_listsessions_cache[profile_name]['sessions']
        else:
            return None

    def list_desktops(self, session_uuid):
        """\
        Retrieve a list of available desktop sessions from the current cache content of
        :class:`x2go.cache.X2GoListSessionsCache` for a given :class:`x2go.session.X2GoSession` instance (specified by its
        unique session UUID).

        :param session_uuid: unique identifier of session to query cache for
        :type session_uuid: ``str``
        :returns: a list of strings representing X2Go desktop sessions available for sharing
        :rtype: ``list`` (or ``None``)

        """
        profile_name = self.client_instance.get_session_profile_name(session_uuid)
        if self.is_cached(session_uuid=session_uuid):
            return self.x2go_listsessions_cache[profile_name]['desktops']
        else:
            return None

    def list_mounts(self, session_uuid):
        """\
        Retrieve a list of mounted client shares from the current cache content of
        :class:`x2go.cache.X2GoListSessionsCache` for a given :class:`x2go.session.X2GoSession` instance (specified by its
        unique session UUID).

        :param session_uuid: unique identifier of session to query cache for
        :type session_uuid: ``str``
        :returns: a list of strings representing mounted client shares
        :rtype: ``list`` (or ``None``)

        """
        profile_name = self.client_instance.get_session_profile_name(session_uuid)
        if self.is_cached(session_uuid=session_uuid):
            return self.x2go_listsessions_cache[profile_name]['mounts']
        else:
            return None

    def is_cached(self, profile_name=None, session_uuid=None, cache_type=None):
        """\
        Check if session information is cached.

        :param profile_name: name of profile to update (Default value = None)
        :type profile_name: ``str``
        :param session_uuid: unique identifier of session to query cache for (Default value = None)
        :type session_uuid: ``str``
        :param cache_type: cache type (e.g. 'mounts', 'desktops', 'sessions') (Default value = None)
        :type cache_type: ``str``
        :returns: ``True`` if session information is cached
        :rtype: ``bool``

        """
        if profile_name is None and session_uuid and self.client_instance:
            try:
                profile_name = self.client_instance.get_session_profile_name(session_uuid)
            except x2go_exceptions.X2GoSessionRegistryException:
                raise x2go_exceptions.X2GoSessionCacheException("requested session UUID is not valid anymore")
        _is_profile_cached = profile_name in self.x2go_listsessions_cache
        _is_cache_type_cached = _is_profile_cached and cache_type in self.x2go_listsessions_cache[profile_name]
        if cache_type is None:
            return _is_profile_cached
        else:
            return _is_cache_type_cached



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2010-2023 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import os

from setuptools import setup, find_packages

# silence pyflakes, the correct __VERSION__ will be detected below...
__VERSION__ = "0.0.0.0"
try:
    # for python3.x
    for line in open(os.path.join('x2go', '__init__.py'),encoding='utf-8').readlines():
        if (line.startswith('__VERSION__')):
            exec(line.strip())
except TypeError:
    # for older python2.x versions
    for line in file(os.path.join('x2go', '__init__.py')).readlines():
        if (line.startswith('__VERSION__')):
            exec(line.strip())
MODULE_VERSION = __VERSION__

setup(
    name = "x2go",
    version = MODULE_VERSION,
    description = "Python X2Go implements an X2Go client/session library in Python based on the Python Paramiko SSH module.",
    license = 'AGPLv3+',
    author = 'Mike Gabriel',
    url = 'http://www.x2go.org',
    packages = find_packages('.'),
    package_dir = {'': '.'},
    install_requires = ['gevent', 'paramiko', ]
)
